@@ -19,7 +19,7 @@ use strict;
 use warnings;
 use Glib;
 
-our $VERSION = '0.041';
+our $VERSION = '0.045';
 
 use Carp;
 $Carp::Internal{(__PACKAGE__)}++;
@@ -527,6 +527,21 @@ example:
   my $button = Gtk3::Button->new;
   Gtk3::WidgetClass::find_style_property ($button, 'image-spacing')
 
+=head2 Overriding virtual functions
+
+When subclassing a gtk+ class or when implementing a gtk+ interface with
+L<Glib::Object::Subclass>, you can override any virtual functions that the
+class has by simply defining sub routines with names obtained by capitalizing
+the original names of the virtual functions.  So, for example, if you implement
+a custom subclass of C<Gtk3::CellRenderer> and want to override its virtual
+function C<render>, you provide a sub routine with the name C<RENDER> in your
+package.
+
+  sub RENDER {
+    my ($cell, $cr, $widget, $background_area, $cell_area, $flags) = @_;
+    # do something
+  }
+
 =head1 DESCRIPTION FOR LIBRARY BINDING AUTHORS
 
 =head2 C<< Glib::Object::Introspection->setup >>
@@ -575,7 +590,7 @@ flattened so that they return plain lists.  For example
     'Gtk3::Window::list_toplevels'
   ]
 
-The function names refer to those after name corrections.  Functions occuring
+The function names refer to those after name corrections.  Functions occurring
 in C<flatten_array_ref_return_for> may also occur in C<class_static_methods>.
 
 =item handle_sentinel_boolean_for => [ function1, ... ]
@@ -592,7 +607,7 @@ be returned, and otherwise an empty list will be returned.
     'Gtk3::TreeSelection::get_selected'
   ]
 
-The function names refer to those after name corrections.  Functions occuring
+The function names refer to those after name corrections.  Functions occurring
 in C<handle_sentinel_boolean_for> may also occur in C<class_static_methods>.
 
 =item use_generic_signal_marshaller_for => [ [package1, signal1, [arg_converter1]], ... ]
@@ -688,18 +703,23 @@ For example:
   $box->child_get_property ($label, 'expand', $gvalue);
   my $value = $gvalue->get_value
 
-=head2 Handling extendable enumerations
+=head2 Handling raw enumerations and flags
 
-If you need to handle extendable enumerations for which more than the
-pre-defined values might be valid, then use C<<
-Glib::Object::Introspection->convert_enum_to_sv >> and C<<
-Glib::Object::Introspection->convert_sv_to_enum >>.  They will raise an
+If you need to handle raw enumerations/flags or extendable enumerations for
+which more than the pre-defined values might be valid, then use C<<
+Glib::Object::Introspection->convert_enum_to_sv >>, C<<
+Glib::Object::Introspection->convert_sv_to_enum >>, C<<
+Glib::Object::Introspection->convert_flags_to_sv >> and C<<
+Glib::Object::Introspection->convert_sv_to_flags >>.  They will raise an
 exception on unknown values; catching it then allows you to implement fallback
 behavior.
 
   Glib::Object::Introspection->convert_enum_to_sv (package, enum_value)
   Glib::Object::Introspection->convert_sv_to_enum (package, sv)
 
+  Glib::Object::Introspection->convert_flags_to_sv (package, flags_value)
+  Glib::Object::Introspection->convert_sv_to_flags (package, sv)
+
 =head1 SEE ALSO
 
 =over
